/*******************************************************************************
***  This file is part of Qt Mocks.                                          ***
***                                                                          ***
***  Copyright (C) 2018, 2019, 2020, 2021                                    ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef QT_MOCKS_MATCHERS_H_
#define QT_MOCKS_MATCHERS_H_

#include <QtGlobal>

#include <memory>
#include <ostream>
#include <sstream>
#include <string>
#include <typeinfo>
#include <tuple>
#include <utility>

#include "qt-mocks_export.h"
#include "Output.h"
#include "Verify.hpp"


namespace qt_mocks
{
namespace common_matchers
{

template<typename ValueType>
class ExpectedValue
{
	ValueType value_;

public:
	explicit ExpectedValue(ValueType value);
	ExpectedValue(const ExpectedValue&) = default;
	ExpectedValue(ExpectedValue&&) noexcept = default;
	~ExpectedValue() = default;
	ExpectedValue& operator=(const ExpectedValue&) = default;
	ExpectedValue& operator=(ExpectedValue&&) noexcept = default;
	const ValueType& value() const;

};
template<typename ValueType>
struct Matcher
{
	using MatcherValueType = ValueType;

	explicit Matcher(const ExpectedValue<ValueType>& value);
	explicit Matcher(std::unique_ptr<ExpectedValue<ValueType>> expected);
	Matcher(const Matcher&);
	Matcher(Matcher&& other) noexcept;
	virtual ~Matcher() = default;
	Matcher& operator=(const Matcher&) = delete;
	Matcher& operator=(Matcher&& other) noexcept;
	virtual bool compare(const ValueType& actual) const =0;
	virtual std::string description(const ValueType& actual) const =0;

	std::unique_ptr<ExpectedValue<ValueType>> expected_;
};
template<typename ValueType>
struct EqualMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;
};
template<>
struct QT_MOCKS_EXPORT EqualMatcher<float> : Matcher<float>
{
	using Matcher<float>::Matcher;

	bool compare(const float& actual) const override;
	std::string description(const float& actual) const override;
};
template<>
struct QT_MOCKS_EXPORT EqualMatcher<double> : Matcher<double>
{
	using Matcher<double>::Matcher;

	bool compare(const double& actual) const override;
	std::string description(const double& actual) const override;
};
template<>
struct QT_MOCKS_EXPORT EqualMatcher<long double> : Matcher<long double>
{
	using Matcher<long double>::Matcher;

	bool compare(const long double& actual) const override;
	std::string description(const long double& actual) const override;
};
template<typename ValueType>
struct NotEqualMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;
};
template<typename ValueType>
struct LessThanMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;
};
template<typename ValueType>
struct LessThanEqualMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;
};
template<typename ValueType>
struct GreaterThanMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;
};
template<typename ValueType>
struct GreaterThanEqualMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;
};
template<typename ClassType, typename MemberVariableType, typename MatcherType>
struct MemberVariableMatcher : Matcher<ClassType>
{
	MemberVariableMatcher(MemberVariableType ClassType::* const member, const MatcherType& matcher);
	bool compare(const ClassType& actual) const override;
	std::string description(const ClassType& actual) const override;

	MemberVariableType ClassType::* const member_;
	MatcherType matcher_;
};
template<typename ClassType, typename MemberFunctionType, typename MatcherType>
struct MemberFunctionMatcher : Matcher<ClassType>
{
	MemberFunctionMatcher(
		MemberFunctionType (ClassType::* member)() const,
		const MatcherType& matcher
	);
	bool compare(const ClassType& actual) const override;
	std::string description(const ClassType& actual) const override;

	MemberFunctionType (ClassType::* member_)() const;
	MatcherType matcher_;
};
template<typename ValueType>
struct BetweenExclusiveMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	BetweenExclusiveMatcher(const ValueType& begin, const ValueType& end);
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	ValueType end_;
};
template<typename ValueType>
struct BetweenBeginInclusiveMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	BetweenBeginInclusiveMatcher(const ValueType& begin, const ValueType& end);
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	ValueType end_;
};
template<typename ValueType>
struct BetweenEndInclusiveMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	BetweenEndInclusiveMatcher(const ValueType& begin, const ValueType& end);
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	ValueType end_;
};
template<typename ValueType>
struct BetweenInclusiveMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	BetweenInclusiveMatcher(const ValueType& begin, const ValueType& end);
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	ValueType end_;
};
template<typename ValueType>
struct NegationMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	template<typename MatcherType>
	explicit NegationMatcher(MatcherType&& matcher);
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	std::unique_ptr<Matcher<ValueType>> matcher_;
};
struct QT_MOCKS_EXPORT BooleanMatcher : Matcher<bool>
{
	explicit BooleanMatcher(const bool expected);
	bool compare(const bool& actual) const override;
	std::string description(const bool& actual) const override;
};
template<typename Container>
struct IsEmptyMatcher : Matcher<Container>
{
	IsEmptyMatcher();
	bool compare(const Container& actual) const override;
	std::string description(const Container& actual) const override;
};
template<typename ValueType, typename ... MatcherTypes>
struct AllOfMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	explicit AllOfMatcher(MatcherTypes&& ... matchers);
	AllOfMatcher(const AllOfMatcher&) = delete;
	AllOfMatcher(AllOfMatcher&&) noexcept;
	~AllOfMatcher() override = default;
	AllOfMatcher& operator=(const AllOfMatcher&) = delete;
	AllOfMatcher& operator=(AllOfMatcher&&) noexcept;
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	std::tuple<MatcherTypes...> matchers_;
};
template<typename ValueType, typename ... MatcherTypes>
struct AnyOfMatcher : Matcher<ValueType>
{
	using Matcher<ValueType>::Matcher;

	explicit AnyOfMatcher(MatcherTypes&& ... matchers);
	AnyOfMatcher(const AnyOfMatcher&) = delete;
	AnyOfMatcher(AnyOfMatcher&&) noexcept;
	~AnyOfMatcher() override = default;
	AnyOfMatcher& operator=(const AnyOfMatcher&) = delete;
	AnyOfMatcher& operator=(AnyOfMatcher&&) noexcept;
	bool compare(const ValueType& actual) const override;
	std::string description(const ValueType& actual) const override;

	std::tuple<MatcherTypes...> matchers_;
};

}

/**
 * @brief A matcher that ensures that all of the supplied matcher objects passed.
 *
 * @tparam ValueType This is the type the match objects will compare against.
 */
template<typename ValueType, typename ... MatcherTypes>
common_matchers::AllOfMatcher<ValueType, MatcherTypes...> allOf(MatcherTypes&& ... matchers);
/**
 * @brief A matcher that ensures that any of the supplied matcher objects passed.
 *
 * @tparam ValueType This is the type the match objects will compare against.
 */
template<typename ValueType, typename ... MatcherTypes>
common_matchers::AnyOfMatcher<ValueType, MatcherTypes...> anyOf(MatcherTypes&& ... matchers);
/**
 * @brief Creates a matcher that a value will be false against it.
 */
QT_MOCKS_EXPORT qt_mocks::common_matchers::BooleanMatcher isFalse();
/**
 * @brief Creates a matcher that a value will be true against it.
 */
QT_MOCKS_EXPORT qt_mocks::common_matchers::BooleanMatcher isTrue();
/**
 * @brief Creates a matcher that will expect that a value is equal to the matcher's expected value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param expectedValue The expected value of the test value.
 */
template<typename ValueType>
qt_mocks::common_matchers::EqualMatcher<ValueType> equals(const ValueType& expected);
/**
 * @brief Creates a matcher that will expect that a value is not equal to the matcher's expected
 * value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param expectedValue The expected value of the test value.
 */
template<typename ValueType>
qt_mocks::common_matchers::NotEqualMatcher<ValueType> notEquals(const ValueType& expected);
/**
 * @brief Creates a matcher that will expect that a value is less-than the matcher's expected value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param expectedValue The expected value of the test value.
 */
template<typename ValueType>
qt_mocks::common_matchers::LessThanMatcher<ValueType> lessThan(const ValueType& expected);
/**
 * @brief Creates a matcher that will expect that a value is less-than or equal to the matcher's
 * expected value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param expectedValue The expected value of the test value.
 */
template<typename ValueType>
qt_mocks::common_matchers::LessThanEqualMatcher<ValueType> lessThanEqual(const ValueType& expected);
/**
 * @brief Creates a matcher that will expect that a value is greater-than the matcher's expected
 * value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param expectedValue The expected value of the test value.
 */
template<typename ValueType>
qt_mocks::common_matchers::GreaterThanMatcher<ValueType> greaterThan(const ValueType& expected);
/**
 * @brief Creates a matcher that will expect that a value is greater-than or equal to the
 * matcher's expected value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param expectedValue The expected value of the test value.
 */
template<typename ValueType>
qt_mocks::common_matchers::GreaterThanEqualMatcher<ValueType> greaterThanEqual(
	const ValueType& expected
);
/**
 * @brief Creates a matcher that will expect that a value is between two values, exclusive, to the
 * test value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param beginValue The expected value must be greater than this value.
 * @param endValue The expected value must be less than this value.
 */
template<typename ValueType>
qt_mocks::common_matchers::BetweenExclusiveMatcher<ValueType> between(
	const ValueType& beginValue,
	const ValueType& endValue
);
/**
 * @brief Creates a matcher that will expect that a value is between two values,
 * beginning inclusive and ending exclusive, to the test value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param beginValue The expected value must be greater than or equal to this value.
 * @param endValue The expected value must be less than this value.
 */
template<typename ValueType>
qt_mocks::common_matchers::BetweenBeginInclusiveMatcher<ValueType> betweenBeginInclusive(
	const ValueType& beginValue,
	const ValueType& endValue
);
/**
 * @brief Creates a matcher that will expect that a value is between two values,
 * beginning exclusive and ending inclusive, to the test value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param beginValue The expected value must be greater than this value.
 * @param endValue The expected value must be less than or equal to this value.
 */
template<typename ValueType>
qt_mocks::common_matchers::BetweenEndInclusiveMatcher<ValueType> betweenEndInclusive(
	const ValueType& beginValue,
	const ValueType& endValue
);
/**
 * @brief Creates a matcher that will expect that a value is between two values,
 * beginning and ending inclusive, to the test value.
 *
 * @tparam ValueType This is the type the match object will compare against.
 *
 * @param beginValue The expected value must be greater than or equal to this value.
 * @param endValue The expected value must be less than or equal to this value.
 */
template<typename ValueType>
qt_mocks::common_matchers::BetweenInclusiveMatcher<ValueType> betweenInclusive(
	const ValueType& beginValue,
	const ValueType& endValue
);
/**
 * @brief Creates a matcher that will negate the result of another matcher.
 *
 * @tparam MatcherType The type of the matcher to be negated.
 *
 * @param matcher The matcher whose comparison result should be negated.
 */
template<typename MatcherType>
qt_mocks::common_matchers::NegationMatcher<typename MatcherType::MatcherValueType> not_(
	MatcherType&& matcher
);
/**
 * @brief Creates a matcher that will expect that an object's member variable meets the conditions
 * of a matcher.
 *
 * @tparam ClassType The type of the class the member points to.
 * @tparam MemberVariableType The type of the member variable.
 * @tparam MatcherType The type of the matcher to be negated.
 *
 * @param memberVariable A pointer to the the member variable contained within the object to use
 * when comparing the actual value, the member variable, to the matcher's expected value.
 * @param matcher The matcher that tests the value of the object's member variable.
 */
template<typename ClassType, typename MemberVariableType, typename MatcherType>
common_matchers::MemberVariableMatcher<ClassType, MemberVariableType, MatcherType> memberVariable(
	MemberVariableType ClassType::* const memberVariable,
	const MatcherType& matcher
);
/**
 * @brief Creates a matcher that will expect that an object's no-argument member function meets the
 * conditions of a matcher. This function is most often the accessor to an object's private or
 * protected member variable.
 *
 * @tparam ClassType The type of the class the member points to.
 * @tparam MemberReturnType The return type of the member function.
 * @tparam MatcherType The type of the matcher to be negated.
 *
 * @param mumberFunction A pointer to the no-argument member function within the object to use to
 * obtain the actual value to compare against the matcher's expected value.
 * @param matcher The matcher that tests the value of the object's member function.
 */
template<typename ClassType, typename MemberReturnType, typename MatcherType>
common_matchers::MemberFunctionMatcher<ClassType, MemberReturnType, MatcherType> memberFunction(
	MemberReturnType (ClassType::* mumberFunction)() const,
	MatcherType matcher
);
/**
 * @brief Creates a matcher for a container that is expected to be empty.
 *
 * @tparam ContainerType The type of the matcher to be negated.
 */
template<typename ContainerType>
common_matchers::IsEmptyMatcher<ContainerType> isEmpty();

namespace internal
{

template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
typename std::enable_if<Index == sizeof...(MatcherTypes), bool>::type compareAll(
	const ValueType&,
	const std::tuple<MatcherTypes...>&
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
typename std::enable_if<(Index < sizeof...(MatcherTypes)), bool>::type compareAll(
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
typename std::enable_if<Index == sizeof...(MatcherTypes), std::string>::type descriptionAll(
	std::ostringstream& oss,
	const ValueType&,
	const std::tuple<MatcherTypes...>&
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
typename std::enable_if<(Index < sizeof...(MatcherTypes)), std::string>::type descriptionAll(
	std::ostringstream& oss,
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
typename std::enable_if<Index == sizeof...(MatcherTypes), bool>::type compareAny(
	const ValueType&,
	const std::tuple<MatcherTypes...>&
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
typename std::enable_if<(Index < sizeof...(MatcherTypes)), bool>::type compareAny(
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
typename std::enable_if<Index == sizeof...(MatcherTypes), std::string>::type descriptionAny(
	std::ostringstream& oss,
	const ValueType&,
	const std::tuple<MatcherTypes...>&
);
template<typename ValueType, std::size_t Index = 0, typename ... MatcherTypes>
typename std::enable_if<(Index < sizeof...(MatcherTypes)), std::string>::type descriptionAny(
	std::ostringstream& oss,
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
);

}}

template<typename ValueType>
qt_mocks::common_matchers::ExpectedValue<ValueType>::ExpectedValue(ValueType value) :
	value_{std::move(value)}
{}

template<typename ValueType>
const ValueType& qt_mocks::common_matchers::ExpectedValue<ValueType>::value() const
{
	return this->value_;
}

template<typename ValueType>
qt_mocks::common_matchers::Matcher<ValueType>::Matcher(
	const ExpectedValue<ValueType>& value
) :
	qt_mocks::common_matchers::Matcher<ValueType>::Matcher{
		std::make_unique<ExpectedValue<ValueType>>(value)
	}
{}

template<typename ValueType>
qt_mocks::common_matchers::Matcher<ValueType>::Matcher(
	std::unique_ptr<ExpectedValue<ValueType>> expected
) :
	expected_{std::move(expected)}
{}

template<typename ValueType>
qt_mocks::common_matchers::Matcher<ValueType>::Matcher(const Matcher& other) :
	expected_{
		other.expected_ == nullptr
			? nullptr
			: std::make_unique<ExpectedValue<ValueType>>(other.expected_->value())
	}
{}

template<typename ValueType>
qt_mocks::common_matchers::Matcher<ValueType>::Matcher(Matcher&& other) noexcept :
	expected_{std::move(other.expected_)}
{}

template<typename ValueType>
qt_mocks::common_matchers::Matcher<ValueType>&
		qt_mocks::common_matchers::Matcher<ValueType>::operator=(Matcher&& other) noexcept
{
	if (this != &other)
	{
		this->expected_ = std::move(other.expected_);
	}
	return *this;
}

template<typename ValueType>
bool qt_mocks::common_matchers::EqualMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual == this->expected_->value();
}


template<typename ValueType>
std::string qt_mocks::common_matchers::EqualMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be equal to the expected value, "
			<< this->expected_->value() << '.';
	return oss.str();
}

template<typename ValueType>
bool qt_mocks::common_matchers::NotEqualMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual != this->expected_->value();
}

template<typename ValueType>
std::string qt_mocks::common_matchers::NotEqualMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should not be equal to the expected value, "
		<< this->expected_->value() << '.';
	return oss.str();
}

template<typename ValueType>
bool qt_mocks::common_matchers::LessThanMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual < this->expected_->value();
}

template<typename ValueType>
std::string qt_mocks::common_matchers::LessThanMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be less than the expected value, "
		<< this->expected_->value() << '.';
	return oss.str();
}

template<typename ValueType>
bool qt_mocks::common_matchers::LessThanEqualMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual <= this->expected_->value();
}

template<typename ValueType>
std::string qt_mocks::common_matchers::LessThanEqualMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be less than or equal to the expected "
		"value, " << this->expected_->value() << '.';
	return oss.str();
}

template<typename ValueType>
bool qt_mocks::common_matchers::GreaterThanMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual > this->expected_->value();
}

template<typename ValueType>
std::string qt_mocks::common_matchers::GreaterThanMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be greater than to the expected value, "
		<< this->expected_->value() << '.';
	return oss.str();
}

template<typename ValueType>
bool qt_mocks::common_matchers::GreaterThanEqualMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual >= this->expected_->value();
}

template<typename ValueType>
std::string qt_mocks::common_matchers::GreaterThanEqualMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be greater than or equal to the expected "
		"value, " << this->expected_->value() << '.';
	return oss.str();
}

template<typename ClassType, typename MemberVariableType, typename MatcherType>
qt_mocks::common_matchers::MemberVariableMatcher<
	ClassType,
	MemberVariableType,
	MatcherType
>::MemberVariableMatcher(
	MemberVariableType ClassType::* const member,
	const MatcherType& matcher
) :
	qt_mocks::common_matchers::Matcher<ClassType>{nullptr},
	member_{member},
	matcher_{matcher}
{}

template<typename ClassType, typename MemberVariableType, typename MatcherType>
bool qt_mocks::common_matchers::MemberVariableMatcher<
	ClassType,
	MemberVariableType,
	MatcherType
>::compare(
	const ClassType& actual
) const
{
	Q_ASSERT(this->expected_ == nullptr);
	return this->matcher_.compare(actual.*this->member_);
}

template<typename ClassType, typename MemberVariableType, typename MatcherType>
std::string qt_mocks::common_matchers::MemberVariableMatcher<
	ClassType,
	MemberVariableType,
	MatcherType
>::description(
	const ClassType& actual
) const
{
	Q_ASSERT(this->expected_ == nullptr);
	return "The member variable with " + this->matcher_.description(actual.*this->member_);
}

template<typename ClassType, typename MemberFunctionType, typename MatcherType>
qt_mocks::common_matchers::MemberFunctionMatcher<
	ClassType,
	MemberFunctionType,
	MatcherType
>::MemberFunctionMatcher(
	MemberFunctionType (ClassType::* member)() const,
	const MatcherType& matcher
) :
	qt_mocks::common_matchers::Matcher<ClassType>{nullptr},
	member_{member},
	matcher_{matcher}
{}

template<typename ClassType, typename MemberFunctionType, typename MatcherType>
bool qt_mocks::common_matchers::MemberFunctionMatcher<
	ClassType,
	MemberFunctionType,
	MatcherType
>::compare(
	const ClassType& actual
) const
{
	Q_ASSERT(this->expected_ == nullptr);
	return this->matcher_.compare((actual.*this->member_)());
}

template<typename ClassType, typename MemberVariableType, typename MatcherType>
std::string qt_mocks::common_matchers::MemberFunctionMatcher<
	ClassType,
	MemberVariableType,
	MatcherType
>::description(
	const ClassType& actual
) const
{
	Q_ASSERT(this->expected_ == nullptr);
	return "The member accessor should return with "
		+ this->matcher_.description((actual.*this->member_)());
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenExclusiveMatcher<ValueType>::BetweenExclusiveMatcher(
	const ValueType& begin,
	const ValueType& end
) :
	qt_mocks::common_matchers::Matcher<ValueType>{
		qt_mocks::common_matchers::ExpectedValue<ValueType>{begin}
	},
	end_{end}
{
	Q_ASSERT(begin <= end);
}

template<typename ValueType>
bool qt_mocks::common_matchers::BetweenExclusiveMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual > this->expected_->value()
		&& actual < this->end_;
}

template<typename ValueType>
std::string qt_mocks::common_matchers::BetweenExclusiveMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be between " << this->expected_->value()
			<< " and " << this->end_ << ", exclusive.";
	return oss.str();
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenBeginInclusiveMatcher<ValueType>::BetweenBeginInclusiveMatcher(
	const ValueType& begin,
	const ValueType& end
) :
	qt_mocks::common_matchers::Matcher<ValueType>{
		qt_mocks::common_matchers::ExpectedValue<ValueType>{begin}
	},
	end_{end}
{
	Q_ASSERT(begin <= end);
}

template<typename ValueType>
bool qt_mocks::common_matchers::BetweenBeginInclusiveMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual >= this->expected_->value()
		&& actual < this->end_;
}

template<typename ValueType>
std::string qt_mocks::common_matchers::BetweenBeginInclusiveMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be between " << this->expected_->value()
			<< " and " << this->end_ << ", beginning inclusive and ending exclusive.";
	return oss.str();
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenEndInclusiveMatcher<ValueType>::BetweenEndInclusiveMatcher(
	const ValueType& begin,
	const ValueType& end
) :
	qt_mocks::common_matchers::Matcher<ValueType>{
		qt_mocks::common_matchers::ExpectedValue<ValueType>{begin}
	},
	end_{end}
{
	Q_ASSERT(begin <= end);
}

template<typename ValueType>
bool qt_mocks::common_matchers::BetweenEndInclusiveMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual > this->expected_->value()
		&& actual <= this->end_;
}

template<typename ValueType>
std::string qt_mocks::common_matchers::BetweenEndInclusiveMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be between " << this->expected_->value()
		<< " and " << this->end_ << ", beginning exclusive and ending inclusive.";
	return oss.str();
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenInclusiveMatcher<ValueType>::BetweenInclusiveMatcher(
	const ValueType& begin,
	const ValueType& end
) :
	qt_mocks::common_matchers::Matcher<ValueType>{
		qt_mocks::common_matchers::ExpectedValue<ValueType>{begin}
	},
	end_{end}
{
	Q_ASSERT(begin <= end);
}

template<typename ValueType>
bool qt_mocks::common_matchers::BetweenInclusiveMatcher<ValueType>::compare(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	return actual >= this->expected_->value()
		&& actual <= this->end_;
}

template<typename ValueType>
std::string qt_mocks::common_matchers::BetweenInclusiveMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	Q_ASSERT(this->expected_ != nullptr);
	std::ostringstream oss;
	oss << "The actual value, " << actual << ", should be between " << this->expected_->value()
		<< " and " << this->end_ << ", beginning and ending inclusive.";
	return oss.str();
}

template<typename ValueType>
template<typename MatcherType>
qt_mocks::common_matchers::NegationMatcher<ValueType>::NegationMatcher(MatcherType&& matcher) :
	qt_mocks::common_matchers::Matcher<ValueType>{nullptr},
	matcher_{std::make_unique<MatcherType>(std::move(matcher))}
{}

template<typename ValueType>
bool qt_mocks::common_matchers::NegationMatcher<ValueType>::compare(const ValueType& actual) const
{
	return ! this->matcher_->compare(actual);
}

template<typename ValueType>
std::string qt_mocks::common_matchers::NegationMatcher<ValueType>::description(
	const ValueType& actual
) const
{
	std::ostringstream oss;
	oss << "Expected the negation of the following: " << this->matcher_->description(actual);
	return oss.str();
}

template<typename Container>
qt_mocks::common_matchers::IsEmptyMatcher<Container>::IsEmptyMatcher() :
	qt_mocks::common_matchers::Matcher<Container>{nullptr}
{}

template<typename Container>
bool qt_mocks::common_matchers::IsEmptyMatcher<Container>::compare(const Container& actual) const
{
	return actual.empty();
}

template<typename Container>
std::string qt_mocks::common_matchers::IsEmptyMatcher<Container>::description(
	const Container& actual
) const
{
	std::ostringstream oss;
	oss << "The actual container is not empty. Contains: ";

	bool isFirst{true};
	for (const auto& value : actual)
	{
		oss << (isFirst ? "" : ", ") << value;
		isFirst = false;
	}
	oss << ')';
	return oss.str();
}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...>::AllOfMatcher(
	MatcherTypes&& ... matchers
) :
	qt_mocks::common_matchers::Matcher<ValueType>{nullptr},
	matchers_{std::make_tuple(std::move(matchers)...)}
{}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...>::AllOfMatcher::AllOfMatcher(
	AllOfMatcher&& other
) noexcept :
	qt_mocks::common_matchers::Matcher<ValueType>{nullptr},
	matchers_{std::move(other.matchers_)}
{}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...>&
		qt_mocks::common_matchers::AllOfMatcher<
			ValueType,
			MatcherTypes...
		>::AllOfMatcher::operator=(
	qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...>&& other
) noexcept
{
	if (this != &other)
		this->matchers_ = std::move(other.matchers_);
	return *this;
}

template<typename ValueType, typename ... MatcherTypes>
bool qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...>::compare(
	const ValueType& actual
) const
{
	return internal::compareAll(actual, this->matchers_);
}

template<typename ValueType, typename ... MatcherTypes>
std::string qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...>::description(
	const ValueType& actual
) const
{
	std::ostringstream oss;
	return internal::descriptionAll(oss, actual, this->matchers_);
}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>::AnyOfMatcher(
	MatcherTypes&& ... matchers
) :
	qt_mocks::common_matchers::Matcher<ValueType>{nullptr},
	matchers_{std::make_tuple(std::move(matchers)...)}
{}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>::AnyOfMatcher::AnyOfMatcher(
	AnyOfMatcher&& other
) noexcept :
	qt_mocks::common_matchers::Matcher<ValueType>{nullptr},
	matchers_{std::move(other.matchers_)}
{}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>&
		qt_mocks::common_matchers::AnyOfMatcher<
			ValueType,
			MatcherTypes...
		>::AnyOfMatcher::operator=(
	qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>&& other
) noexcept
{
	if (this != &other)
		this->matchers_ = std::move(other.matchers_);
	return *this;
}

template<typename ValueType, typename ... MatcherTypes>
bool qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>::compare(
	const ValueType& actual
) const
{
	return internal::compareAny(actual, this->matchers_);
}

template<typename ValueType, typename ... MatcherTypes>
std::string qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>::description(
	const ValueType& actual
) const
{
	std::ostringstream oss;
	return internal::descriptionAny(oss, actual, this->matchers_);
}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AllOfMatcher<ValueType, MatcherTypes...> qt_mocks::allOf(
	MatcherTypes&& ... matchers
)
{
	return common_matchers::AllOfMatcher<ValueType, MatcherTypes...>{
		std::forward<MatcherTypes>(matchers)...
	};
}

template<typename ValueType, typename ... MatcherTypes>
qt_mocks::common_matchers::AnyOfMatcher<ValueType, MatcherTypes...> qt_mocks::anyOf(
	MatcherTypes&& ... matchers
)
{
	return common_matchers::AnyOfMatcher<ValueType, MatcherTypes...>{
		std::forward<MatcherTypes>(matchers)...
	};
}

template<typename ValueType>
qt_mocks::common_matchers::EqualMatcher<ValueType> qt_mocks::equals(const ValueType& expected)
{
	return common_matchers::EqualMatcher<ValueType>{
		common_matchers::ExpectedValue<ValueType>{expected}
	};
}

template<typename ValueType>
qt_mocks::common_matchers::NotEqualMatcher<ValueType> qt_mocks::notEquals(const ValueType& expected)
{
	return common_matchers::NotEqualMatcher<ValueType>{
		common_matchers::ExpectedValue<ValueType>{expected}
	};
}

template<typename ValueType>
qt_mocks::common_matchers::LessThanMatcher<ValueType> qt_mocks::lessThan(const ValueType& expected)
{
	return common_matchers::LessThanMatcher<ValueType>{
		common_matchers::ExpectedValue<ValueType>{expected}
	};
}

template<typename ValueType>
qt_mocks::common_matchers::LessThanEqualMatcher<ValueType> qt_mocks::lessThanEqual(
	const ValueType& expected
)
{
	return common_matchers::LessThanEqualMatcher<ValueType>{
		common_matchers::ExpectedValue<ValueType>{expected}
	};
}

template<typename ValueType>
qt_mocks::common_matchers::GreaterThanMatcher<ValueType> qt_mocks::greaterThan(
	const ValueType& expected
)
{
	return common_matchers::GreaterThanMatcher<ValueType>{
		common_matchers::ExpectedValue<ValueType>{expected}
	};
}

template<typename ValueType>
qt_mocks::common_matchers::GreaterThanEqualMatcher<ValueType> qt_mocks::greaterThanEqual(
	const ValueType& expected
)
{
	return common_matchers::GreaterThanEqualMatcher<ValueType>{
		common_matchers::ExpectedValue<ValueType>{expected}
	};
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenExclusiveMatcher<ValueType> qt_mocks::between(
	const ValueType& beginValue,
	const ValueType& endValue
)
{
	return common_matchers::BetweenExclusiveMatcher<ValueType>{beginValue, endValue};
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenBeginInclusiveMatcher<ValueType> qt_mocks::betweenBeginInclusive(
	const ValueType& beginValue,
	const ValueType& endValue
)
{
	return common_matchers::BetweenBeginInclusiveMatcher<ValueType>{beginValue, endValue};
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenEndInclusiveMatcher<ValueType> qt_mocks::betweenEndInclusive(
	const ValueType& beginValue,
	const ValueType& endValue
)
{
	return common_matchers::BetweenEndInclusiveMatcher<ValueType>{beginValue, endValue};
}

template<typename ValueType>
qt_mocks::common_matchers::BetweenInclusiveMatcher<ValueType> qt_mocks::betweenInclusive(
	const ValueType& beginValue,
	const ValueType& endValue
)
{
	return common_matchers::BetweenInclusiveMatcher<ValueType>{beginValue, endValue};
}

template<typename MatcherType>
qt_mocks::common_matchers::NegationMatcher<
	typename MatcherType::MatcherValueType
> qt_mocks::not_(
	MatcherType&& matcher
)
{
	return common_matchers::NegationMatcher<typename MatcherType::MatcherValueType>{
		std::move(matcher)
	};
}

template<typename ClassType, typename MemberVariableType, typename MatcherType>
qt_mocks::common_matchers::MemberVariableMatcher<ClassType, MemberVariableType, MatcherType>
		qt_mocks::memberVariable(
	MemberVariableType ClassType::* const memberVariable,
	const MatcherType& matcher
)
{
	static_assert(
		std::is_member_object_pointer<decltype(memberVariable)>::value,
		"The supplied memberVariable must be a pointer to a class member variable."
	);
	return common_matchers::MemberVariableMatcher<ClassType, MemberVariableType, MatcherType>{
		memberVariable,
		matcher
	};
}

template<typename ClassType, typename MemberReturnType, typename MatcherType>
qt_mocks::common_matchers::MemberFunctionMatcher<ClassType, MemberReturnType, MatcherType>
		qt_mocks::memberFunction(
	MemberReturnType (ClassType::* mumberFunction)() const,
	MatcherType matcher
)
{
	static_assert(
		std::is_member_function_pointer<decltype(mumberFunction)>::value,
		"The supplied mumberFunction must be a pointer to a class member function."
	);
	return common_matchers::MemberFunctionMatcher<ClassType, MemberReturnType, decltype(matcher)>{
		mumberFunction,
		matcher
	};
}

template<typename ContainerType>
qt_mocks::common_matchers::IsEmptyMatcher<ContainerType> qt_mocks::isEmpty()
{
	return common_matchers::IsEmptyMatcher<ContainerType>{};
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
typename std::enable_if<Index == sizeof...(MatcherTypes), bool>::type
		qt_mocks::internal::compareAll(
	const ValueType&,
	const std::tuple<MatcherTypes...>&
)
{
	return true;
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
typename std::enable_if<(Index < sizeof...(MatcherTypes)), bool>::type
		qt_mocks::internal::compareAll(
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
)
{
	if ( ! std::get<Index>(matchers).compare(actual))
		return false;
	return compareAll<ValueType, Index + 1>(actual, matchers);
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
typename std::enable_if<Index == sizeof...(MatcherTypes), std::string>::type
		qt_mocks::internal::descriptionAll(
	std::ostringstream& oss,
	const ValueType&,
	const std::tuple<MatcherTypes...>&
)
{
	return oss.str();
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
typename std::enable_if<(Index < sizeof...(MatcherTypes)), std::string>::type
		qt_mocks::internal::descriptionAll(
	std::ostringstream& oss,
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
)
{
#ifdef _WIN32
#	pragma warning(push)
#	// Justification: C++ 14 doesn't support constexpr if
#	pragma warning(disable: 4127)
#endif
	if (Index == 0)
		oss << "Expected all of the following to be true. ";
#ifdef _WIN32
#	pragma warning(pop)
#endif
	const auto& matcher{std::get<Index>(matchers)};
	const bool matchFailed{ ! matcher.compare(actual)};
	oss << (matchFailed ? "FAILED: " : "") << matcher.description(actual);

#ifdef _WIN32
#	pragma warning(push)
#	// Justification: C++ 14 doesn't support constexpr if
#	pragma warning(disable: 4127)
#endif
	if (Index < sizeof...(MatcherTypes) - 1)
		oss << ' ';
#ifdef _WIN32
#	pragma warning(pop)
#endif
	return descriptionAll<ValueType, Index + 1>(oss, actual, matchers);
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
typename std::enable_if<Index == sizeof...(MatcherTypes), bool>::type
		qt_mocks::internal::compareAny(
	const ValueType&,
	const std::tuple<MatcherTypes...>&
)
{
	return false;
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
typename std::enable_if<(Index < sizeof...(MatcherTypes)), bool>::type
		qt_mocks::internal::compareAny(
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
)
{
	if (std::get<Index>(matchers).compare(actual))
		return true;
	return compareAny<ValueType, Index + 1>(actual, matchers);
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
typename std::enable_if<Index == sizeof...(MatcherTypes), std::string>::type
		qt_mocks::internal::descriptionAny(
	std::ostringstream& oss,
	const ValueType&,
	const std::tuple<MatcherTypes...>&
)
{
	return oss.str();
}

template<typename ValueType, std::size_t Index, typename ... MatcherTypes>
typename std::enable_if<(Index < sizeof...(MatcherTypes)), std::string>::type
		qt_mocks::internal::descriptionAny(
	std::ostringstream& oss,
	const ValueType& actual,
	const std::tuple<MatcherTypes...>& matchers
)
{
#ifdef _WIN32
#	pragma warning(push)
#	// Justification: C++ 14 doesn't support constexpr if
#	pragma warning(disable: 4127)
#endif
	if (Index == 0)
		oss << "Expected any of the following to be true. ";
#ifdef _WIN32
#	pragma warning(pop)
#endif
	const auto& matcher{std::get<Index>(matchers)};
	const bool matchFailed{ ! matcher.compare(actual)};
	oss << (matchFailed ? "FAILED: " : "") << matcher.description(actual);

#ifdef _WIN32
#	pragma warning(push)
#	// Justification: C++ 14 doesn't support constexpr if
#	pragma warning(disable: 4127)
#endif
	if (Index < sizeof...(MatcherTypes) - 1)
		oss << ' ';
#ifdef _WIN32
#	pragma warning(pop)
#endif
	return descriptionAny<ValueType, Index + 1>(oss, actual, matchers);
}

#endif
