/*******************************************************************************
***  This file is part of Qt Mocks.                                          ***
***                                                                          ***
***  Copyright (C) 2018, 2019, 2020, 2021                                    ***
***  John Schneiderman <Licencing _AT_ Schneiderman _DOT_ me>                ***
***                                                                          ***
***  This program is free software: you can redistribute it and/or modify it ***
***  under the terms of the GNU Lesser General Public License as published   ***
***  by the Free Software Foundation, either version 3 of the License, or    ***
***  (at your option) any later version.                                     ***
***                                                                          ***
***  This program is distributed in the hope that it will be useful, but     ***
***  WITHOUT ANY WARRANTY; without even the implied warranty of              ***
***  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    ***
***  See the GNU Lesser General Public License for more details.             ***
***                                                                          ***
***  You should have received a copy of the GNU Lesser General Public License***
***  along with this program. If not, see <http://www.gnu.org/licenses/>.    ***
*******************************************************************************/
#ifndef QT_MOCKS_VERIFY_HPP
#define QT_MOCKS_VERIFY_HPP

#include <QString>
#include <QTest>


/**
 * @brief Creates an expectation that a matcher will compare to be true against.
 *
 * @param actualValueGenerator The actual value to compare against the expectation.
 * @param matcher The matcher that performs the comparison.
 */
#define QVERIFY_THAT(actualValueGenerator, matcher)  \
	do  \
	{  \
		const auto actualValue{actualValueGenerator};  \
		\
		if (  \
			 ! QTest::qVerify(  \
				matcher.compare(actualValue),  \
				"Verification of " #actualValueGenerator,  \
				matcher.description(actualValue).c_str(),  \
				__FILE__,  \
				__LINE__  \
			)  \
		)  \
			return;  \
	}  \
	while (false)
/**
 * @brief Creates an expectation that a value will be false against it.
 *
 * @param actualValue The actual value to compare.
 */
#define QVERIFY_FALSE(actualValue) QVERIFY_THAT(actualValue, qt_mocks::isFalse())
/**
 * @brief Creates an expectation that a value will be true against it.
 *
 * @param actualValue The actual value to compare.
 */
#define QVERIFY_TRUE(actualValue) QVERIFY_THAT(actualValue, qt_mocks::isTrue())
/**
 * @brief Confirms that a function invocation does not throw an exception.
 *
 * @param functionCall The function to invoke and test.
 */
#define QVERIFY_NO_THROW(functionCall)  \
	do  \
		try  \
		{  \
			functionCall;  \
		}  \
		catch (const std::exception& error)  \
		{  \
			const QString message{  \
				"Expected '" #functionCall "' to not throw, but throws exception type '"  \
					+ QString{typeid(error).name()} + "'. Error message: " + error.what()  \
			};  \
			QFAIL(message.toUtf8());  \
		}  \
		catch (...)  \
		{  \
			QFAIL("Unknown non-standard exception thrown by '" #functionCall "'.");  \
		}  \
	while(false)

#endif
